#include <QDebug>
#include <QtXml/qdom.h>
#include <QFile>
#include <QFileDialog>
#include "../common/sources/system/cgraphics2d.h"
#include "../common/sources/system/filecache.h"
#include "interface.h"
#include "game.h"
#include "player.h"
#include "keyboard.h"
#include "objects.h"
#include "spells.h"
#include "attacks.h"
#include "tiles.h"
#include "monsters.h"

CInterface   interface;

static const int   championsColors[4][3] =
{
    {  0, 222,   0},
    {255, 255,   0},
    {255,   0,   0},
    {  0,   0, 255}
};

//------------------------------------------------------------
CInterface::CInterface()
{
    mainState = eMainGame;
    weaponsAreaState = eWeaponsAreaWeapons;
    currentChampion = -1;
    currentWeapon = -1;
    currentSpellCaster = 0;
    isResurrecting = false;
    isPressingEye = false;
    selectedChampion = 0;
    shouldLoad = false;
    shouldSave = false;

    arrowsRects[eArrowTurnLeft]  = CRect(CVec2(234, 125), CVec2(261, 145));
    arrowsRects[eArrowForward]   = CRect(CVec2(263, 125), CVec2(289, 145));
    arrowsRects[eArrowTurnRight] = CRect(CVec2(291, 125), CVec2(318, 145));

    arrowsRects[eArrowLeft]      = CRect(CVec2(234, 147), CVec2(261, 167));
    arrowsRects[eArrowBackward]  = CRect(CVec2(263, 147), CVec2(289, 167));
    arrowsRects[eArrowRight]     = CRect(CVec2(291, 147), CVec2(318, 167));

    // initialize the positions of objects in the character's sheet
    bodyInfos[CCharacter::eBodyPartHead].pos      = CVec2(34,  59);
    bodyInfos[CCharacter::eBodyPartNeck].pos      = CVec2( 6,  66);
    bodyInfos[CCharacter::eBodyPartTorso].pos     = CVec2(34,  79);
    bodyInfos[CCharacter::eBodyPartLeftHand].pos  = CVec2( 6,  86);
    bodyInfos[CCharacter::eBodyPartRightHand].pos = CVec2(62,  86);
    bodyInfos[CCharacter::eBodyPartLegs].pos      = CVec2(34,  99);
    bodyInfos[CCharacter::eBodyPartFeet].pos      = CVec2(34, 119);
    bodyInfos[CCharacter::eBodyPartPouch1].pos    = CVec2( 6, 106);
    bodyInfos[CCharacter::eBodyPartPouch2].pos    = CVec2( 6, 123);
    bodyInfos[CCharacter::eBodyPartQuiver1].pos   = CVec2(62, 106);
    bodyInfos[CCharacter::eBodyPartQuiver2].pos   = CVec2(79, 106);
    bodyInfos[CCharacter::eBodyPartQuiver3].pos   = CVec2(62, 123);
    bodyInfos[CCharacter::eBodyPartQuiver4].pos   = CVec2(79, 123);

    bodyInfos[CCharacter::eBodyPartHead].type      = 'H';
    bodyInfos[CCharacter::eBodyPartNeck].type      = 'N';
    bodyInfos[CCharacter::eBodyPartTorso].type     = 'T';
    bodyInfos[CCharacter::eBodyPartLeftHand].type  = 'h';
    bodyInfos[CCharacter::eBodyPartRightHand].type = 'h';
    bodyInfos[CCharacter::eBodyPartLegs].type      = 'L';
    bodyInfos[CCharacter::eBodyPartFeet].type      = 'F';
    bodyInfos[CCharacter::eBodyPartPouch1].type    = 'P';
    bodyInfos[CCharacter::eBodyPartPouch2].type    = 'P';
    bodyInfos[CCharacter::eBodyPartQuiver1].type   = 'Q';
    bodyInfos[CCharacter::eBodyPartQuiver2].type   = 'q';
    bodyInfos[CCharacter::eBodyPartQuiver3].type   = 'q';
    bodyInfos[CCharacter::eBodyPartQuiver4].type   = 'q';

    for (int i = 0; i < BACKPACK_SIZE; ++i)
    {
        int x = 66 + ((i + 1) % 9) * 17;
        int y = 49 + ((i + 1) / 9) * 17;
        bodyInfos[CCharacter::eBodyPartBackpack + i].pos = CVec2(x, y);
        bodyInfos[CCharacter::eBodyPartBackpack + i].type = 'B';
    }

    // default images for the body parts
    for (int i = 0; i < CCharacter::eBodyPartCount; ++i)
        bodyInfos[i].defaultImage = -1;

    bodyInfos[CCharacter::eBodyPartHead].defaultImage         = 24;
    bodyInfos[CCharacter::eBodyPartNeck].defaultImage         = 16;
    bodyInfos[CCharacter::eBodyPartTorso].defaultImage        = 26;
    bodyInfos[CCharacter::eBodyPartLeftHand].defaultImage     = 20;
    bodyInfos[CCharacter::eBodyPartRightHand].defaultImage    = 22;
    bodyInfos[CCharacter::eBodyPartLegs].defaultImage         = 28;
    bodyInfos[CCharacter::eBodyPartFeet].defaultImage         = 30;
    bodyInfos[CCharacter::eBodyPartPouch1].defaultImage       = 17;
    bodyInfos[CCharacter::eBodyPartQuiver1].defaultImage      = 18;
    bodyInfos[CCharacter::eBodyPartBackpack + 8].defaultImage = 19;
}

//------------------------------------------------------------
CInterface::~CInterface()
{
}

//------------------------------------------------------------
void    CInterface::display(QImage* image)
{
    if (isInInventory() == true)
        drawInventory(image);
    else if (mainState == eMainDiskDialog)
        drawDiskDialog(image);
    else if (mainState == eMainLoadingDialog)
        drawLoadingDialog(image);
    else if (mainState == eMainSavingDialog)
        drawSavingDialog(image);
    else if (mainState == eMainLoadedDialog)
        drawLoadedDialog(image);
    else if (mainState == eMainSavedDialog)
        drawSavedDialog(image);
    else if (mainState == eMainQuitConfirm)
        drawQuitConfirm(image);
    else if (mainState == eMainPause)
        drawPause(image);
    else if (mainState == eMainSleep)
        drawSleep(image);

    for (int i = 0; i < 4; ++i)
        drawChampion(image, i);

    drawObjectName(image);
    drawSpells(image);
    drawWeaponsArea(image);
    drawArrows(image);
    drawMessages(image);
}

//------------------------------------------------------------
bool    CInterface::isInInventory()
{
    return (mainState == eMainInventory);
}

//------------------------------------------------------------
bool    CInterface::isInDiskDialog()
{
    return (mainState == eMainDiskDialog ||
            mainState == eMainLoadingDialog ||
            mainState == eMainSavingDialog ||
            mainState == eMainLoadedDialog ||
            mainState == eMainSavedDialog ||
            mainState == eMainQuitConfirm);
}

//------------------------------------------------------------
bool    CInterface::isChampionEmpty(int num)
{
    CCharacter* c = &game.characters[num];
    return c->portrait == -1;
}

//------------------------------------------------------------
bool    CInterface::isChampionDead(int num)
{
    CCharacter* c = &game.characters[num];
    return c->isDead();
}

//------------------------------------------------------------
bool    CInterface::isWeaponEmpty(int num)
{
    return (isChampionDead(num) || isChampionEmpty(num));
}

//------------------------------------------------------------
bool    CInterface::isWeaponGreyed(int num)
{
    return (weaponCoolDown[num].isRunning() == true ||
            isResurrecting ||
            isGamePaused() == true ||
            isSleeping() == true ||
            isFalling() == true);
}

//------------------------------------------------------------
bool    CInterface::isArrowsGreyed()
{
    return (isInInventory() ||
            isGamePaused() == true ||
            isSleeping() == true ||
            isFalling() == true);
}

//------------------------------------------------------------
bool    CInterface::isSpellsGreyed()
{
    return (isResurrecting ||
            isGamePaused() == true ||
            isSleeping() == true ||
            isFalling() == true);
}

//------------------------------------------------------------
void    CInterface::drawBar(QImage* image, int num, CVec2 pos, CCharacter::EStatsNames statName)
{
    CCharacter::SStat   stat = getStat(num, statName);

    if  (stat.maxValue == 0)
        return;

    pos.y = 27 - (25 * stat.value) / stat.maxValue;

    if (pos.y < 2)
        pos.y = 2;

    if (pos.y <= 26)
    {
        CRect   rect(pos, CVec2(pos.x + 3, 26));
        QColor color(championsColors[num][0],
                     championsColors[num][1],
                     championsColors[num][2]);
        graph2D.rect(image, rect, color, true);
    }
}

//------------------------------------------------------------
void    CInterface::drawChampionName(QImage* image, int num, CVec2 pos, bool isFullName)
{
    CCharacter* c = &game.characters[num];
    std::string name = c->firstName;

    if (isFullName == true)
        name += " " + c->lastName;

    QColor  color = (num == selectedChampion ? HIGHLIGHT_TEXT_COLOR : DEFAULT_TEXT_COLOR);
    drawText(image, pos, eFontStandard, name.c_str(), color);
}

//------------------------------------------------------------
void    CInterface::drawChampion(QImage* image, int num)
{
    if (isChampionEmpty(num) == false)
    {
        CVec2   basePos(69 * num, 0);
        CVec2   namePos = basePos + CVec2(1, 1);
        CRect   handMouseRect(basePos + CVec2(0, 7),  basePos + CVec2(42, 28));
        CRect   barsMouseRect(basePos + CVec2(43, 0), basePos + CVec2(66, 28));
        CCharacter* c = &game.characters[num];

        if (isChampionDead(num) == true)
        {
            QImage  champBg = fileCache.getImage("gfx/interface/DeadChampInfo.png");
            graph2D.drawImage(image, basePos, champBg);

            // draw name
            drawChampionName(image, num, namePos, false);
        }
        else
        {
            QImage  champBg = fileCache.getImage("gfx/interface/ChampInfo.png");
            graph2D.drawImage(image, basePos, champBg);

            if ((isInInventory() == true || isInDiskDialog() == true) && currentChampion == num)
            {
                // draw a background rectangle
                CRect   rect2(basePos, basePos + CVec2(42, 6));
                graph2D.rect(image, rect2, INVENTORY_BG_COLOR, true);

                CVec2   headPos = basePos + CVec2(7, 0);
                CVec2   headSize = CVec2(CHAMP_PORTRAIT_WIDTH - 1, CHAMP_PORTRAIT_HEIGHT - 1);
                CRect   headRect(headPos, headPos + headSize);
                graph2D.rect(image, headRect, INVENTORY_SLOT_COLOR, true);

                // draw head
                CRect   imageRect = getChampionPortraitRect(c->portrait);
                QImage  portraits = fileCache.getImage("gfx/interface/ChampPortraits.png");
                graph2D.drawImageAtlas(image, headPos, portraits, imageRect);

                if (c->damagesTimer.isRunning() == true)
                {
                    QImage  damagesImage = fileCache.getImage("gfx/interface/DamageReceivedBig.png");
                    graph2D.drawImage(image, basePos + CVec2(6, 0), damagesImage);
                    static char damagesStr[8];
                    sprintf(damagesStr, "%d", c->displayedDamages);
                    int     length = strlen(damagesStr);
                    CVec2   pos(20 - (length - 1) * 3, 12);
                    drawText(image, basePos + pos, eFontStandard, damagesStr, DAMAGES_COLOR);
                }

                // mouse area
                if (isResurrecting == false || isInDiskDialog() == true)
                    mouse.addArea(eMouseArea_3DView, barsMouseRect, eCursor_Hand);
            }
            else
            {
                // draw rectangles
                CRect   rect(basePos + CVec2(3, 9),
                             basePos + CVec2(20, 26));
                graph2D.rect(image, rect, INVENTORY_SLOT_COLOR, false);

                CRect   rect2(basePos + CVec2(23, 9),
                              basePos + CVec2(40, 26));
                graph2D.rect(image, rect2, INVENTORY_SLOT_COLOR, false);

                // draw hands
                mouse.addArea(eMouseArea_None, handMouseRect, eCursor_Hand);
                bool    activateHandsArea = (isResurrecting == false && isGamePaused() == false && isSleeping() == false && isFalling() == false);
                drawBodyPart(image, basePos + CVec2(4, 10), num, CCharacter::eBodyPartLeftHand, activateHandsArea);
                drawBodyPart(image, basePos + CVec2(24, 10), num, CCharacter::eBodyPartRightHand, activateHandsArea);

                // draw name
                if (c->damagesTimer.isRunning() == false)
                {
                    drawChampionName(image, num, namePos, false);
                }
                else
                {
                    QImage  damagesImage = fileCache.getImage("gfx/interface/DamageReceivedSmall.png");
                    graph2D.drawImage(image, basePos, damagesImage);
                    static char damagesStr[8];
                    sprintf(damagesStr, "%d", c->displayedDamages);
                    int     length = strlen(damagesStr);
                    CVec2   pos(19 - (length - 1) * 3, 1);
                    drawText(image, basePos + pos, eFontStandard, damagesStr, DAMAGES_COLOR);
                }

                // mouse area
                if (activateHandsArea == true)
                    mouse.addArea(eMouseArea_Inventory, barsMouseRect, eCursor_Hand, (void*)num);

                CRect nameRect(basePos, basePos + CVec2(42, 6));
                mouse.addArea(eMouseArea_SelectChamp, nameRect, eCursor_Arrow, (void*)num);
            }

            // bars
            drawBar(image, num, basePos + CVec2(46, 2), CCharacter::eStatHealth);
            drawBar(image, num, basePos + CVec2(53, 2), CCharacter::eStatStamina);
            drawBar(image, num, basePos + CVec2(60, 2), CCharacter::eStatMana);
        }
    }
}

//------------------------------------------------------------
void CInterface::drawObjectName(QImage* image)
{
    int type = mouse.mObjectInHand.getType();

    if (type != 0)
    {
        std::string name;

        if (type == OBJECT_TYPE_BONES)
            name = setChampionNameInString(mouse.mObjectInHand.getIntParam("Champion"), "ITEM149");
        else
            name = objects.mObjectInfos[type].name;

        drawText(image, CVec2(233, 33), eFontStandard, name.c_str(), MAIN_INTERFACE_COLOR);
    }
}

//------------------------------------------------------------
void    CInterface::drawSpells(QImage* image)
{
    // background image
    QImage  spellsBg = fileCache.getImage("gfx/interface/SpellsArea.png");
    graph2D.drawImage(image, CVec2(233, 50), spellsBg);

    // casters names
    CCharacter* c = &game.characters[currentSpellCaster];

    for (int i = 0; i < currentSpellCaster; ++i)
    {
        if (isChampionEmpty(i) == false && isChampionDead(i) == false)
        {
            CRect   rect(CVec2(233 + 14 * i, 42),
                         CVec2(244 + 14 * i, 48));
            graph2D.rect(image, rect, MAIN_INTERFACE_COLOR, true);

            if (isSpellsGreyed() == false)
                mouse.addArea(eMouseArea_SpellCaster, rect, eCursor_Arrow, (void*)i);
        }
    }

    if (isChampionEmpty(currentSpellCaster) == false && isChampionDead(currentSpellCaster) == false)
    {
        CRect   rect(CVec2(233 + 14 * currentSpellCaster, 42),
                     CVec2(277 + 14 * currentSpellCaster, 49));
        graph2D.rect(image, rect, MAIN_INTERFACE_COLOR, true);

        drawText(image, rect.tl + CVec2(2, 2), eFontStandard, c->firstName.c_str(), BLACK);
    }

    for (int i = currentSpellCaster + 1; i < 4; ++i)
    {
        if (isChampionEmpty(i) == false && isChampionDead(i) == false)
        {
            CRect   rect(CVec2(266 + 14 * i, 42),
                         CVec2(277 + 14 * i, 48));
            graph2D.rect(image, rect, MAIN_INTERFACE_COLOR, true);

            if (isSpellsGreyed() == false)
                mouse.addArea(eMouseArea_SpellCaster, rect, eCursor_Arrow, (void*)i);
        }
    }

    // spells characters
    int currentPage = 0;

    if (isChampionEmpty(currentSpellCaster) == false && isChampionDead(currentSpellCaster) == false)
        currentPage = strlen(c->spell) % 4;

    char word[2] = "a";

    for (int i = 0; i < 6; ++i)
    {
        CVec2 pos(239 + i * 14, 54);
        word[0] = 'a' + currentPage * 6 + i;
        drawText(image, pos, eFontStandard, word, MAIN_INTERFACE_COLOR);

        if (isChampionEmpty(currentSpellCaster) == false &&
            isChampionDead(currentSpellCaster) == false &&
            isSpellsGreyed() == false)
        {
            CRect   rect(CVec2(235 + 14 * i, 51),
                         CVec2(247 + 14 * i, 61));
            mouse.addArea(eMouseArea_SpellLetter, rect, eCursor_Arrow, (void*)((int)word[0]));
        }
    }

    // current spell
    if (isChampionEmpty(currentSpellCaster) == false && isChampionDead(currentSpellCaster) == false)
    {
        drawText(image, CVec2(237, 66), eFontStandard, c->spell, MAIN_INTERFACE_COLOR, 1);

        if (isSpellsGreyed() == false)
        {
            CRect   rect(CVec2(234, 63), CVec2(303, 73));
            mouse.addArea(eMouseArea_CastSpell, rect, eCursor_Arrow);
        }
    }

    // back arrow
    if (isChampionEmpty(currentSpellCaster) == false && isChampionDead(currentSpellCaster) == false)
    {
        CRect   rect(CVec2(305, 63), CVec2(318, 73));
        mouse.addArea(eMouseArea_SpellBack, rect, eCursor_Arrow);
    }

    // grey out
    if (isSpellsGreyed() == true)
    {
        CRect   rect(CVec2(233, 42),
                     CVec2(319, 74));
        graph2D.patternRectangle(image, rect);
    }
}

//------------------------------------------------------------
void    CInterface::drawWeaponsArea(QImage* image)
{
    if (weaponsAreaState == eWeaponsAreaWeapons)
    {
        for (int i = 0; i < 4; ++i)
            drawWeapon(image, i);
    }
    else if (weaponsAreaState == eWeaponsAreaAttacks)
    {
        drawAttacks(image, currentWeapon);
    }
    else
    {
        drawDamages(image);
    }
}

//------------------------------------------------------------
void    CInterface::drawWeapon(QImage* image, int num)
{
    if (isWeaponEmpty(num) == false)
    {
        // draw the background rectangle
        CRect   rect(CVec2(233 + 22 * num, 86),
                     CVec2(252 + 22 * num, 120));
        graph2D.rect(image, rect, MAIN_INTERFACE_COLOR, true);

        // draw the weapon
        CVec2       pos = CVec2(235 + 22 * num, 96);
        CObject&    weapon = getWeapon(num);
        int         weaponType = weapon.getType();
        CObjects::CObjectInfo   weaponInfo = objects.mObjectInfos[weaponType];

        if (weaponInfo.combo != 0)
        {
            std::string imageFile;
            int         imageNum;

            CObjects::CObjectInfo  object = objects.mObjectInfos[weaponType];
            imageFile = object.imageFile.toLocal8Bit().constData();
            imageNum = object.imageNum;

            QImage  objImage = fileCache.getImage(imageFile);
            CRect   objRect = getItemRect(imageNum);
            graph2D.drawImageAtlas(image, pos, objImage, objRect, BLACK);
        }

        // grey out or mouse area
        if (isWeaponGreyed(num) == true)
            graph2D.patternRectangle(image, rect);
        else if (weaponInfo.combo != 0)
            mouse.addArea(eMouseArea_Weapon, rect, eCursor_Arrow, (void*)num);
    }
}

//------------------------------------------------------------
void    CInterface::drawAttacks(QImage* image, int num)
{
    CObject&    weapon = getWeapon(num);
    int         weaponType = weapon.getType();
    CObjects::CObjectInfo   weaponInfo = objects.mObjectInfos[weaponType];
    CAttacks::SCombo&       combo = attacks.getCombo(weaponInfo.combo);

    // draw the background image
    QImage  attacksBg = fileCache.getImage("gfx/interface/WeaponsActions.png");
    CVec2   pos(233, 77);
    graph2D.drawImage(image, pos, attacksBg);

    // hide unused slots
    for (int i = 0; i < 3; ++i)
    {
        if (combo.attacks[i].type == 0)
        {
            CRect   rect(CVec2(233, 86 + i * 12),
                         CVec2(319, 97 + i * 12));
            graph2D.rect(image, rect, BLACK, true);
        }
    }

    // draw character's name
    CCharacter* c = &game.characters[num];
    drawText(image, CVec2(235, 79), eFontStandard, c->firstName.c_str(), BLACK);

    // draw attacks' names
    for (int i = 0; i < 3; ++i)
    {
        if (combo.attacks[i].type != 0)
        {
            static char attackName[16];
            sprintf(attackName, "ATTACK%02d", combo.attacks[i].type);
            drawText(image, CVec2(241, 89 + i * 12), eFontStandard, getText(attackName).c_str(), MAIN_INTERFACE_COLOR);

            if (attackHighlightTime.isRunning() == true && currentAttack == i)
            {
                CRect   rect(CVec2(234, 86 + i * 12),
                             CVec2(318, 96 + i * 12));
                graph2D.Xor(image, rect, MAIN_INTERFACE_COLOR);
            }
        }
    }

    // mouse areas
    if (isWeaponGreyed(num) == false && attackHighlightTime.isRunning() == false)
    {
        for (int i = 0; i < 3; ++i)
        {
            if (combo.attacks[i].type != 0)
            {
                CRect   rect(CVec2(234, 86 + i * 12),
                             CVec2(318, 96 + i * 12));
                mouse.addArea(eMouseArea_Attack, rect, eCursor_Arrow, (void*)i);
            }
        }

        CRect   rect(CVec2(290, 77), CVec2(314, 83));
        mouse.addArea(eMouseArea_CloseAttack, rect, eCursor_Arrow);
    }
}

//------------------------------------------------------------
void CInterface::drawDamages(QImage *image)
{
    QImage  damagesBg = fileCache.getImage("gfx/interface/DamageDone.png");
    static char damagesStr[8];
    sprintf(damagesStr, "%d", damages);

    int     length = strlen(damagesStr);
    float   scaleX = 0.44f + (length - 1) * 0.24f;
    CVec2   pos(258 - (length - 1) * 10, 81);

    graph2D.drawImageScaled(image, pos, damagesBg, scaleX, false, 0.82f);

    CVec2   textPos(274 - (length - 1) * 3, 97);
    drawText(image, textPos, eFontStandard, damagesStr, MAIN_INTERFACE_COLOR);
}

//------------------------------------------------------------
void    CInterface::drawArrows(QImage* image)
{
    QImage  arrowsBg = fileCache.getImage("gfx/interface/MovementArrows.png");
    CVec2   pos(233, 124);
    graph2D.drawImage(image, pos, arrowsBg);

    if (isArrowsGreyed() == true)
    {
        CRect   rect(CVec2(233, 124),
                     CVec2(319, 168));
        graph2D.patternRectangle(image, rect);
    }
    else
    {
        static const EMouseAreaTypes arrowsMouseAreas[] =
        {
            eMouseArea_ArrowForward,
            eMouseArea_ArrowBackward,
            eMouseArea_ArrowLeft,
            eMouseArea_ArrowRight,
            eMouseArea_ArrowTurnLeft,
            eMouseArea_ArrowTurnRight
        };

        for (int i = 0; i < eArrowsCount; ++i)
        {
            mouse.addArea(arrowsMouseAreas[i], arrowsRects[i]);
        }

        if (arrowHighlightTime.isRunning() == true)
            graph2D.Xor(image, arrowsRects[currentArrow], MAIN_INTERFACE_COLOR);
    }
}

//------------------------------------------------------------
void    CInterface::displayStat(QImage* image, CVec2 pos, int x2, CCharacter::EStatsNames num)
{
    static char temp[256];

    sprintf(temp, "STATS%02d", num);
    drawText(image, pos, eFontStandard, getText(temp).c_str(), DEFAULT_TEXT_COLOR);

    CCharacter::SStat stat = getStat(currentChampion, num);
    sprintf(temp, "%3d/%3d", stat.value, stat.maxValue);
    pos.x = x2;
    drawText(image, pos, eFontStandard, temp, DEFAULT_TEXT_COLOR);
}

//------------------------------------------------------------
void    CInterface::drawScroll(QImage* image, CObject& object)
{
    QImage  scrollBg = fileCache.getImage("gfx/interface/Scroll.png");
    graph2D.drawImage(image, CVec2(80, 85), scrollBg);

    if (isPressingEye == true)
    {
        QImage  eye = fileCache.getImage("gfx/interface/Eye.png");
        graph2D.drawImage(image, CVec2(83, 90), eye);
    }
    else
    {
        QImage  arrow = fileCache.getImage("gfx/interface/Arrow.png");
        graph2D.drawImage(image, CVec2(83, 90), arrow);
    }

    std::string text = object.getStringParam("Text");
    drawTextMultiLines(image, CVec2(162, 91), eFontScroll, getText(text), SCROLL_TEXT_COLOR, 59, 7);
}

//------------------------------------------------------------
void    CInterface::drawInfos(QImage* image)
{
    int objType = mouse.mObjectInHand.getType();

    if (objType == 0)
    {
        // character's infos
        QImage  infoBg = fileCache.getImage("gfx/interface/EmptyInfo.png");
        graph2D.drawImage(image, CVec2(80, 85), infoBg);

        displayStat(image, CVec2(108, 115), 174, CCharacter::eStatStrength);
        displayStat(image, CVec2(108, 122), 174, CCharacter::eStatDexterity);
        displayStat(image, CVec2(108, 129), 174, CCharacter::eStatWisdom);
        displayStat(image, CVec2(108, 136), 174, CCharacter::eStatVitality);
        displayStat(image, CVec2(108, 143), 174, CCharacter::eStatAntiMagic);
        displayStat(image, CVec2(108, 150), 174, CCharacter::eStatAntiFire);
    }
    else if (objType == OBJECT_TYPE_SCROLL)
    {
        // scroll
        drawScroll(image, mouse.mObjectInHand);
    }
    else
    {
        // object's infos
        QImage  infoBg = fileCache.getImage("gfx/interface/EmptyInfo.png");
        graph2D.drawImage(image, CVec2(80, 85), infoBg);

        QImage  eye = fileCache.getImage("gfx/interface/Eye.png");
        graph2D.drawImage(image, CVec2(83, 90), eye);

        QImage  circle = fileCache.getImage("gfx/interface/Circle.png");
        graph2D.drawImage(image, CVec2(106, 87), circle);

        CObjects::CObjectInfo&  infos = objects.mObjectInfos[objType];
        QImage  objSheet = fileCache.getImage(infos.imageFile.toLocal8Bit().constData());
        CRect   objRect = interface.getItemRect(infos.imageNum);
        graph2D.drawImageAtlas(image, CVec2(111, 92), objSheet, objRect);

        // object name
        drawText(image, CVec2(134, 98), eFontStandard, infos.name.c_str(), DEFAULT_TEXT_COLOR);

        int y = 116;

        if (objType == 144)
        {
            // waterskin full
            drawText(image, CVec2(108, y), eFontStandard, getText("OBJ_INFO00").c_str(), DEFAULT_TEXT_COLOR);
            y += 7;
        }
        else if (objType == 145)
        {
            // waterskin empty
            drawText(image, CVec2(108, y), eFontStandard, getText("OBJ_INFO03").c_str(), DEFAULT_TEXT_COLOR);
            y += 7;
        }
        else if (infos.positions.contains("c"))
        {
            // consumable
            drawText(image, CVec2(108, y), eFontStandard, getText("OBJ_INFO04").c_str(), DEFAULT_TEXT_COLOR);
            y += 7;
        }

        // weight
        static char temp[256];
        sprintf(temp, "%s %.1f KG.", getText("OBJ_INFO05").c_str(), (float)infos.weight / 10.0f);
        drawText(image, CVec2(108, y), eFontStandard, temp, DEFAULT_TEXT_COLOR);
    }
}

//------------------------------------------------------------
void    CInterface::drawInventory(QImage* image)
{
    CRect   mainRect(CVec2(0, 33), CVec2(MAINVIEW_WIDTH - 1, 33 + MAINVIEW_HEIGHT - 1));
    mouse.addArea(eMouseArea_None, mainRect, eCursor_Hand);

    QImage  bodyParts = fileCache.getImage("gfx/interface/Items6.png");

    // draw background
    QImage  inventoryBg = fileCache.getImage("gfx/interface/InventoryMain.png");
    graph2D.drawImage(image, CVec2(0, 33), inventoryBg);

    // display name
    drawChampionName(image, currentChampion, CVec2(3, 36), true);

    // display health/stamina/mana
    displayStat(image, CVec2(5, 145), 55, CCharacter::eStatHealth);
    displayStat(image, CVec2(5, 153), 55, CCharacter::eStatStamina);
    displayStat(image, CVec2(5, 161), 55, CCharacter::eStatMana);

    // display load
    CCharacter* c = &game.characters[currentChampion];
    QColor  loadColor;
    if (c->overload() == 0)
        loadColor = DEFAULT_TEXT_COLOR;
    else if (c->overload() == 1)
        loadColor = OVERLOAD1_COLOR;
    else
        loadColor = OVERLOAD2_COLOR;

    drawText(image, CVec2(104, 161), eFontStandard, getText("STATS09").c_str(), loadColor);
    static char temp[256];
    CCharacter::SStat stat = getStat(currentChampion, CCharacter::eStatLoad);
    sprintf(temp, "%5.1f/%3d KG", (float)stat.value / 10.0f, stat.maxValue / 10);

    drawText(image, CVec2(148, 161), eFontStandard, temp, loadColor);

    // eye
    CRect   eyePosRect(CVec2(12, 46), CVec2(27, 61));
    mouse.addArea(eMouseArea_Eye, eyePosRect, (mouse.mButtonPressed == false ? eCursor_Hand : eCursor_None));

    if (isPressingEye == true)
    {
        CRect   imageRect = getItemRect(11);
        graph2D.drawImageAtlas(image, eyePosRect.tl, bodyParts, imageRect);
    }

    // display objects
    for (int i = 0; i < CCharacter::eBodyPartCount; ++i)
        drawBodyPart(image, bodyInfos[i].pos, currentChampion, (CCharacter::EBodyParts)i);

    // buttons
    if (isResurrecting == true)
    {
        // hide save/sleep/close buttons
        CRect   rect(CVec2(174, 35), CVec2(218, 45));
        graph2D.rect(image, rect, INVENTORY_BG_COLOR, true);
    }
    else
    {
        // disk button
        CRect   diskRect(CVec2(174, 36), CVec2(182, 44));
        mouse.addArea(eMouseArea_DiskMenu, diskRect, eCursor_Hand);

        // sleep button
        CRect   sleepRect(CVec2(187, 35), CVec2(205, 45));
        mouse.addArea(eMouseArea_Sleep, sleepRect, eCursor_Hand);

        // close button
        CRect   closeRect(CVec2(210, 36), CVec2(218, 44));
        mouse.addArea(eMouseArea_3DView, closeRect, eCursor_Hand);
    }

    if (isPressingEye == true)
    {
        drawInfos(image);
    }
    else if (isResurrecting == true)
    {
        // draw resurrect/reincarnate image
        QImage  resurrectImage = fileCache.getImage("gfx/interface/ResurrectReincarnate.png");
        CVec2   pos(80, 85);
        graph2D.drawImage(image, pos, resurrectImage);

        if (mouse.mObjectInHand.getType() == 0)
        {
            // resurrect mouse area
            CRect   resurrectRect(CVec2(103, 85), CVec2(159, 143));
            mouse.addArea(eMouseArea_Resurrect, resurrectRect, eCursor_Hand);

            // cancel mouse area
            CRect   cancelRect(CVec2(103, 145), CVec2(218, 157));
            mouse.addArea(eMouseArea_3DView, cancelRect, eCursor_Hand);
        }
    }
    else
    {
        // objects in hand
        CObject&    obj = c->bodyObjects[CCharacter::eBodyPartRightHand];

        if (obj.getType() == OBJECT_TYPE_SCROLL)
            drawScroll(image, obj);
    }
}

//------------------------------------------------------------
void    CInterface::drawSmallButton(QImage* image, CVec2 pos, std::string text, EMouseAreaTypes id)
{
    QImage  button;

    // button image
    if (diskButtonTime.isRunning() == false || clickedDiskButton != id)
        button = fileCache.getImage("gfx/interface/DiskButtonSmallUp.png");
    else
        button = fileCache.getImage("gfx/interface/DiskButtonSmallDown.png");
    graph2D.drawImage(image, pos, button);

    // mouse area
    if (diskButtonTime.isRunning() == false)
    {
        CRect   rect(pos, pos + CVec2(97, 26));
        mouse.addArea(id, rect, eCursor_Arrow);
    }

    // text
    int centerX = pos.x + 49;
    CVec2   textPos(centerX - (text.size() * getCharSpacing(eFontStandard)) / 2, pos.y + 11);
    drawText(image, textPos, eFontStandard, text.c_str(), DISK_TEXT_COLOR);
}

//------------------------------------------------------------
void    CInterface::drawBigButton(QImage* image, CVec2 pos, std::string text, EMouseAreaTypes id)
{
    QImage  button;

    // button image
    if (diskButtonTime.isRunning() == false || clickedDiskButton != id)
        button = fileCache.getImage("gfx/interface/DiskButtonBigUp.png");
    else
        button = fileCache.getImage("gfx/interface/DiskButtonBigDown.png");
    graph2D.drawImage(image, pos, button);

    // mouse area
    if (diskButtonTime.isRunning() == false && text.size() != 0)
    {
        CRect   rect(pos, pos + CVec2(203, 26));
        mouse.addArea(id, rect, eCursor_Arrow);
    }

    // text
    int centerX = pos.x + 102;
    CVec2   textPos(centerX - (text.size() * getCharSpacing(eFontStandard)) / 2, pos.y + 11);
    drawText(image, textPos, eFontStandard, text.c_str(), DISK_TEXT_COLOR);
}

//------------------------------------------------------------
void CInterface::drawDiskBackground(QImage* image)
{
    CRect   mainRect(CVec2(0, 33), CVec2(MAINVIEW_WIDTH - 1, 33 + MAINVIEW_HEIGHT - 1));
    mouse.addArea(eMouseArea_None, mainRect, eCursor_Arrow);

    QImage  dialogBg = fileCache.getImage("gfx/interface/DiskDialogBg.png");
    graph2D.drawImage(image, CVec2(0, 33), dialogBg);
}

//------------------------------------------------------------
void CInterface::drawDiskBackground2(QImage* image)
{
    CRect   mainRect(CVec2(0, 33), CVec2(MAINVIEW_WIDTH - 1, 33 + MAINVIEW_HEIGHT - 1));
    mouse.addArea(eMouseArea_None, mainRect, eCursor_Arrow);

    QImage  dialogBg = fileCache.getImage("gfx/interface/DiskDialogBg2.png");
    graph2D.drawImage(image, CVec2(0, 33), dialogBg);
}

//------------------------------------------------------------
void CInterface::drawDiskDialog(QImage* image)
{
    drawDiskBackground(image);

    // buttons
    drawSmallButton(image, CVec2( 10,  95), getText("DISK_MENU00"), eMouseArea_Save);
    drawSmallButton(image, CVec2( 10, 132), getText("DISK_MENU01"), eMouseArea_Load);
    drawSmallButton(image, CVec2(117,  95), getText("DISK_MENU02"), eMouseArea_Quit);
    drawSmallButton(image, CVec2(117, 132), getText("DISK_MENU03"), eMouseArea_QuitDiskMenu);
}

//------------------------------------------------------------
void CInterface::drawLoadingDialog(QImage* image)
{
    drawDiskBackground2(image);
    drawBigButton(image, CVec2(10, 132), "", eMouseArea_None);

    // text
    std::string text = getText("DISK_MENU05");
    CVec2   pos(112 - (text.size() * getCharSpacing(eFontStandard)) / 2, 80);
    drawText(image, pos, eFontStandard, text.c_str(), DISK_TEXT_COLOR);

    if (dummyDiskTimer.update() == true)
        shouldLoad = true;
}

//------------------------------------------------------------
void CInterface::drawSavingDialog(QImage* image)
{
    drawDiskBackground2(image);
    drawBigButton(image, CVec2(10, 132), "", eMouseArea_None);

    // text
    std::string text = getText("DISK_MENU04");
    CVec2   pos(112 - (text.size() * getCharSpacing(eFontStandard)) / 2, 80);
    drawText(image, pos, eFontStandard, text.c_str(), DISK_TEXT_COLOR);

    if (dummyDiskTimer.update() == true)
        shouldSave = true;
}

//------------------------------------------------------------
void CInterface::drawLoadedDialog(QImage* image)
{
    drawDiskBackground2(image);
    drawBigButton(image, CVec2(10, 132), getText("DISK_MENU08"), eMouseArea_GameLoaded);

    // text
    std::string text = getText("DISK_MENU07");
    CVec2   pos(112 - (text.size() * getCharSpacing(eFontStandard)) / 2, 80);
    drawText(image, pos, eFontStandard, text.c_str(), DISK_TEXT_COLOR);
}

//------------------------------------------------------------
void CInterface::drawSavedDialog(QImage* image)
{
    drawDiskBackground2(image);
    drawBigButton(image, CVec2(10, 132), getText("DISK_MENU08"), eMouseArea_GameSaved);

    // text
    std::string text = getText("DISK_MENU06");
    CVec2   pos(112 - (text.size() * getCharSpacing(eFontStandard)) / 2, 80);
    drawText(image, pos, eFontStandard, text.c_str(), DISK_TEXT_COLOR);
}

//------------------------------------------------------------
void CInterface::drawQuitConfirm(QImage* image)
{
    drawDiskBackground(image);
    drawBigButton(image, CVec2(10, 95), getText("DISK_MENU02"), eMouseArea_AcceptQuitConfirm);
    drawBigButton(image, CVec2(10, 132), getText("DISK_MENU03"), eMouseArea_CancelQuitConfirm);

    // text
    drawTextMultiLines(image, CVec2(112, 44), eFontStandard, getText("DISK_MENU09"), DISK_TEXT_COLOR, 40, 8);
}

//------------------------------------------------------------
void CInterface::drawPause(QImage* image)
{
    CRect   mainRect(CVec2(0, 33), CVec2(MAINVIEW_WIDTH - 1, 33 + MAINVIEW_HEIGHT - 1));
    mouse.addArea(eMouseArea_ExitPause, mainRect, eCursor_Arrow);

    std::string text = getText("MISC_TEXT00");
    CVec2   pos(112 - (text.size() * getCharSpacing(eFontStandard)) / 2, 98);
    drawText(image, pos, eFontStandard, text.c_str(), MAIN_INTERFACE_COLOR);
}

//------------------------------------------------------------
void CInterface::drawSleep(QImage* image)
{
    CRect   mainRect(CVec2(0, 33), CVec2(MAINVIEW_WIDTH - 1, 33 + MAINVIEW_HEIGHT - 1));
    mouse.addArea(eMouseArea_WakeUp, mainRect, eCursor_Arrow);

    std::string text = getText("MISC_TEXT01");
    CVec2   pos(112 - (text.size() * getCharSpacing(eFontStandard)) / 2, 98);
    drawText(image, pos, eFontStandard, text.c_str(), MAIN_INTERFACE_COLOR);
}

//------------------------------------------------------------
CObject& CInterface::getWeapon(int num)
{
    CCharacter* c = &game.characters[num];
    return c->bodyObjects[CCharacter::eBodyPartRightHand];
}

//------------------------------------------------------------
CCharacter::SStat   CInterface::getStat(int num, CCharacter::EStatsNames statName)
{
    CCharacter* c = &game.characters[num];
    return c->stats[statName];
}

//------------------------------------------------------------
CRect   CInterface::getChampionPortraitRect(int num)
{
    int x = num % 8;
    int y = num / 8;
    CRect   rect;

    rect.tl.x = x * CHAMP_PORTRAIT_WIDTH;
    rect.tl.y = y * CHAMP_PORTRAIT_HEIGHT;
    rect.br.x = rect.tl.x + CHAMP_PORTRAIT_WIDTH - 1;
    rect.br.y = rect.tl.y + CHAMP_PORTRAIT_HEIGHT - 1;

    return rect;
}

//------------------------------------------------------------
CRect   CInterface::getItemRect(int num)
{
    int x = num % 16;
    int y = num / 16;
    CRect   rect;

    rect.tl.x = x * ITEM_WIDTH;
    rect.tl.y = y * ITEM_HEIGHT;
    rect.br.x = rect.tl.x + ITEM_WIDTH - 1;
    rect.br.y = rect.tl.y + ITEM_HEIGHT - 1;

    return rect;
}

//------------------------------------------------------------
void    CInterface::drawBodyPart(QImage* image, CVec2 pos, int championNum, CCharacter::EBodyParts part, bool enableArea)
{
    CCharacter* c = &game.characters[championNum];
    CObject*    obj = &c->bodyObjects[part];
    int         objType = obj->getType();

    if (objType != 0)
    {
        // draw the object in this slot
        CObjects::CObjectInfo  object = objects.mObjectInfos[objType];
        QImage  objImage = fileCache.getImage(object.imageFile.toLocal8Bit().constData());
        int imageNum = object.imageNum;

        // objects opened in hand
        if (part == CCharacter::eBodyPartRightHand)
        {
            if (objType == OBJECT_TYPE_SCROLL)
                imageNum = 30;
        }

        // torch
        if (part == CCharacter::eBodyPartRightHand ||
            part == CCharacter::eBodyPartLeftHand)
        {
            if (objType == OBJECT_TYPE_TORCH)
            {
                int charges = obj->getIntParam("Charges");

                if (charges > 0)
                    imageNum = 5 + ((charges - 1) * 3) / TORCH_MAX_CHARGES;
            }
        }

        CRect   rect = getItemRect(imageNum);
        graph2D.drawImageAtlas(image, pos, objImage, rect);
    }
    else if (bodyInfos[part].defaultImage != -1)
    {
        // draw the default image
        QImage  bodyParts = fileCache.getImage("gfx/interface/Items6.png");
        CRect   rect = getItemRect(bodyInfos[part].defaultImage);
        graph2D.drawImageAtlas(image, pos, bodyParts, rect);
    }

    // mouse area
    int     objHandType = mouse.mObjectInHand.getType();
    CRect   mouseRect(pos, pos + CVec2(ITEM_WIDTH - 1, ITEM_HEIGHT - 1));
    bool    isPositionCompatible = true;

    // check if the object we are holding can be put in this type of slot
    if (objHandType != 0)
    {
        QString&    positions = objects.mObjectInfos[objHandType].positions;
        isPositionCompatible = positions.contains(bodyInfos[part].type);
    }

    if (isPositionCompatible == true && enableArea == true)
        mouse.addArea(eMouseArea_InvObject, mouseRect, eCursor_Hand, (void*)championNum, (void*)part);
}

//------------------------------------------------------------
int CInterface::getCharSpacing(EFonts font)
{
    if (font == eFontStandard)
        return 6;
    else if (font == eFontScroll)
        return 6;
    else
        return 8;
}

//------------------------------------------------------------
#define CHAR_WIDTH  8

void   CInterface::drawText(QImage* image, CVec2 pos, EFonts font, const char* text, QColor color, int addSpace)
{
    // Sprite number tables for each character in ASCII from char 32 to char 120
    // The lowercase letters are used for the spells
    //  !"#$%&'()*+,-./
    // 0123456789:;<=>?
    // @ABCDEFGHIJKLMNO
    // PQRSTUVWXYZ[\]^_
    // `abcdefghijklmno
    // pqrstuvwx

    // eFontStandard
    static const int codesStandard[] =
    {
        //      !     "     #     $     %     &     '     (     )     *     +     ,     -     .     /
        0x00, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
        //0     1     2     3     4     5     6     7     8     9     :     ;     <     =     >     ?
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
        //@     A     B     C     D     E     F     G     H     I     J     K     L     M     N     O
        0x00, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F,
        //P     Q     R     S     T     U     V     W     X     Y     Z     [     \     ]     ^     _
        0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
        //`     a     b     c     d     e     f     g     h     i     j     k     l     m     n     o
        0x00, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E,
        //p     q     r     s     t     u     v     w     x
        0x6F, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77
    };

    // eFontScroll
    static const int codesScroll[] =
    {
        //      !     "     #     $     %     &     '     (     )     *     +     ,     -     .     /
        0x00, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x1D, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F,
        //0     1     2     3     4     5     6     7     8     9     :     ;     <     =     >     ?
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F,
        //@     A     B     C     D     E     F     G     H     I     J     K     L     M     N     O
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
        //P     Q     R     S     T     U     V     W     X     Y     Z     [     \     ]     ^     _
        0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x5B, 0x5C, 0x5D, 0x5E, 0x5F,
        //`     a     b     c     d     e     f     g     h     i     j     k     l     m     n     o
        0x00, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x6B, 0x6C, 0x6D, 0x6E,
        //p     q     r     s     t     u     v     w     x
        0x6F, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77
    };

    // eFontWalls
    static const int codesWalls[] =
    {
        //      !     "     #     $     %     &     '     (     )     *     +     ,     -     .     /
        0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x21, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1B, 0x1A,
        //0     1     2     3     4     5     6     7     8     9     :     ;     <     =     >     ?
        0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A,
        //@     A     B     C     D     E     F     G     H     I     J     K     L     M     N     O
        0x1A, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E,
        //P     Q     R     S     T     U     V     W     X     Y     Z     [     \     ]     ^     _
        0x0F, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A,
        //`     a     b     c     d     e     f     g     h     i     j     k     l     m     n     o
        0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A,
        //p     q     r     s     t     u     v     w     x
        0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A, 0x1A
    };

    const int* codes;
    int     spacing = getCharSpacing(font) + addSpace;
    QImage  file;

    if (font == eFontStandard)
    {
        codes = codesStandard;
        file = fileCache.getImage("gfx/interface/Font.png");
        pos.x -= 3;
    }
    else if (font == eFontScroll)
    {
        codes = codesScroll;
        file = fileCache.getImage("gfx/interface/Font.png");
        pos.x -= 3;
    }
    else
    {
        codes = codesWalls;
        file = fileCache.getImage("gfx/3DView/WallFont.png");
    }

    int i = 0;

    while (text[i] != 0)
    {
        int c = codes[text[i] - 32];

        CRect   rect(CVec2(c * CHAR_WIDTH, 0),
                     CVec2((c + 1) * CHAR_WIDTH - 1, file.height() - 1));
        CVec2   cPos(pos.x + i * spacing, pos.y);

        graph2D.drawImageAtlas(image, cPos, file, rect, (font != eFontWalls ? color : TRANSPARENT));
        i++;
    }
}

//------------------------------------------------------------
// split a text using '_' as a delimiter
std::vector<std::string> CInterface::getTextLines(std::string text)
{
    std::vector<std::string>    lines;
    size_t  start = 0;
    size_t  end = 0;

    while (true)
    {
        end = text.find('_', start);

        if (end != std::string::npos)
        {
            lines.push_back(text.substr(start, end - start));
            start = end + 1;
        }
        else
        {
            lines.push_back(text.substr(start, end));
            break;
        }
    }
    return lines;
}

//------------------------------------------------------------
void   CInterface::drawTextMultiLines(QImage* image, CVec2 pos, EFonts font, std::string text, QColor color, int height, int lineHeight)
{
    std::vector<std::string> lines = getTextLines(text);

    // center the text vertically
    if (height != 0)
        pos.y += (height - lines.size() * lineHeight) / 2;

    // draw each line of text
    for (size_t i = 0; i < lines.size(); ++i)
    {
        CVec2   pos2 = pos;
        pos2.x -= lines[i].size() * getCharSpacing(font) / 2;
        drawText(image, pos2, font, lines[i].c_str(), color);
        pos.y += lineHeight;

        if (font == eFontWalls && i == 1)
            pos.y += 3;
    }
}

//------------------------------------------------------------
bool   CInterface::isClickingOnArrow(SMouseArea* clickedArea, EMouseAreaTypes type)
{
    return (mouse.mButtonPressing && clickedArea != NULL && clickedArea->type == type);
}

//------------------------------------------------------------
void    CInterface::setDelayedArrow(EArrows arrow)
{
    delayedArrow = arrow;

    int maxOverload = 0;
    for (int i = 0; i < 4; ++i)
    {
        int overload = game.characters[i].overload();

        if (overload > maxOverload)
            maxOverload = overload;
    }

    if (maxOverload == 0)
        delayedArrowTime.set(OVERLOAD0_DELAY, true);
    else if (maxOverload == 1)
        delayedArrowTime.set(OVERLOAD1_DELAY, true);
    else
        delayedArrowTime.set(OVERLOAD2_DELAY, true);
}

//------------------------------------------------------------
void   CInterface::updateArrows(SMouseArea* clickedArea)
{
    arrowHighlightTime.update();

    if (delayedArrowTime.isRunning() == false &&
        isGamePaused() == false &&
        isArrowsGreyed() == false)
    {
        if (isClickingOnArrow(clickedArea, eMouseArea_ArrowForward) == true || keyboard.keys[CKeyboard::eForward] == true)
        {
            arrowHighlightTime.set(ARROWS_HIGHLIGHT_TIME, false);
            currentArrow = eArrowForward;
            setDelayedArrow(eArrowForward);
        }
        else if (isClickingOnArrow(clickedArea, eMouseArea_ArrowBackward) == true || keyboard.keys[CKeyboard::eBackward] == true)
        {
            arrowHighlightTime.set(ARROWS_HIGHLIGHT_TIME, false);
            currentArrow = eArrowBackward;
            setDelayedArrow(eArrowBackward);
        }
        else if (isClickingOnArrow(clickedArea, eMouseArea_ArrowLeft) == true || keyboard.keys[CKeyboard::eLeft] == true)
        {
            arrowHighlightTime.set(ARROWS_HIGHLIGHT_TIME, false);
            currentArrow = eArrowLeft;
            setDelayedArrow(eArrowLeft);
        }
        else if (isClickingOnArrow(clickedArea, eMouseArea_ArrowRight) == true || keyboard.keys[CKeyboard::eRight] == true)
        {
            arrowHighlightTime.set(ARROWS_HIGHLIGHT_TIME, false);
            currentArrow = eArrowRight;
            setDelayedArrow(eArrowRight);
        }
        else if (isClickingOnArrow(clickedArea, eMouseArea_ArrowTurnLeft) == true || keyboard.keys[CKeyboard::eTurnLeft] == true)
        {
            arrowHighlightTime.set(ARROWS_HIGHLIGHT_TIME, false);
            currentArrow = eArrowTurnLeft;
            player.turnLeft();
        }
        else if (isClickingOnArrow(clickedArea, eMouseArea_ArrowTurnRight) == true || keyboard.keys[CKeyboard::eTurnRight] == true)
        {
            arrowHighlightTime.set(ARROWS_HIGHLIGHT_TIME, false);
            currentArrow = eArrowTurnRight;
            player.turnRight();
        }
    }

    // delayed movements
    if (delayedArrowTime.update() == true)
    {
        switch(delayedArrow)
        {
            case eArrowForward:
                player.moveForward();
                break;

            case eArrowBackward:
                player.moveBackward();
                break;

            case eArrowLeft:
                player.moveLeft();
                break;

            case eArrowRight:
                player.moveRight();
                break;

            default:
                break;
        }
    }
}

//------------------------------------------------------------
int CInterface::findLastChampion()
{
    for (int i = 3; i >= 0; --i)
        if (game.characters[i].portrait != -1)
            return i;

    return -1;
}

//------------------------------------------------------------
std::string CInterface::setChampionNameInString(int champion, std::string stringId)
{
    std::string result = getText(stringId);
    size_t  pos = result.find('#');

    if (pos != std::string::npos)
        result.replace(pos, 1, game.characters[champion].firstName);

    return result;
}

//------------------------------------------------------------
void   CInterface::update(SMouseArea* clickedArea)
{
    isPressingEye = false;

    if (clickedArea != NULL)
    {
        if (mouse.mButtonPressing == true)
        {
            switch (clickedArea->type)
            {
                // enter inventory
                case eMouseArea_Inventory:
                    mainState = eMainInventory;
                    currentChampion = (int)clickedArea->param1;
                break;

                // enter 3D view
                case eMouseArea_3DView:
                {
                    if (isResurrecting == true)
                    {
                        int i = findLastChampion();
                        game.characters[i].portrait = -1;
                    }
                    isResurrecting = false;
                    mainState = eMainGame;
                }
                break;

                // resurrect button
                case eMouseArea_Resurrect:
                {
                    game.lastMirrorClicked->setBoolParam("isEmpty", true);
                    isResurrecting = false;
                    mainState = eMainGame;
                    int champ = findLastChampion();
                    std::string message = setChampionNameInString(champ, "MESSAGE00");
                    addMessage(message, champ);
                }
                break;

                // object slot in inventory
                case eMouseArea_InvObject:
                {
                    // get current character and position in inventory
                    int champNum = (int)clickedArea->param1;
                    CCharacter* c = &game.characters[champNum];
                    int objectPos = (int)clickedArea->param2;

                    // swap the object we are holding with the one in the inventory
                    CObject temp = mouse.mObjectInHand;
                    mouse.mObjectInHand = c->bodyObjects[objectPos];
                    c->bodyObjects[objectPos] = temp;

                    // update the loads of the characters
                    c->updateLoad();
                    c = &game.characters[selectedChampion];
                    c->updateLoad();
                }
                break;

                // click on a champion's name
                case eMouseArea_SelectChamp:
                    selectedChampion = (int)clickedArea->param1;
                break;

                // change the spell caster
                case eMouseArea_SpellCaster:
                    currentSpellCaster = (int)clickedArea->param1;
                break;

                // add a symbol to the spell
                case eMouseArea_SpellLetter:
                {
                    CCharacter* c = &game.characters[currentSpellCaster];

                    if (strlen(c->spell) == 4)
                        c->spell[0] = 0;

                    CCharacter::SStat*  mana = &c->stats[CCharacter::eStatMana];
                    int symbol = (int)clickedArea->param1;
                    int symbolMana = spells.getMana(currentSpellCaster, symbol);

                    if (mana->value >= symbolMana)
                    {
                        int i = strlen(c->spell);
                        c->spell[i] = symbol;
                        c->spell[i + 1] = 0;
                        mana->value -= symbolMana;
                    }
                }
                break;

                // spell backspace
                case eMouseArea_SpellBack:
                {
                    CCharacter* c = &game.characters[currentSpellCaster];
                    int i = strlen(c->spell);

                    if (i != 0)
                        c->spell[i - 1] = 0;
                }
                break;

                // cast a spell
                case eMouseArea_CastSpell:
                {
                    CCharacter* c = &game.characters[currentSpellCaster];

                    if (strlen(c->spell) != 0)
                    {
                        spells.cast(currentSpellCaster);
                        c->spell[0] = 0;
                    }
                }
                break;

                // click on a weapon
                case eMouseArea_Weapon:
                    currentWeapon = (int)clickedArea->param1;
                    weaponsAreaState = eWeaponsAreaAttacks;
                break;

                // click on an attack
                case eMouseArea_Attack:
                    currentAttack = (int)clickedArea->param1;
                    attackHighlightTime.set(ATTACK_HIGHLIGHT_TIME, false);
                    sound->play(player.pos, "sound/Combat.wav");
                break;

                // close the attack box
                case eMouseArea_CloseAttack:
                    weaponsAreaState = eWeaponsAreaWeapons;
                break;

                // click on the disk menu
                case eMouseArea_DiskMenu:
                    mainState = eMainDiskDialog;
                break;

                // disk buttons
                case eMouseArea_Save:
                case eMouseArea_Load:
                case eMouseArea_Quit:
                case eMouseArea_QuitDiskMenu:
                case eMouseArea_GameLoaded:
                case eMouseArea_GameSaved:
                case eMouseArea_AcceptQuitConfirm:
                case eMouseArea_CancelQuitConfirm:
                    clickedDiskButton = clickedArea->type;
                    diskButtonTime.set(DISK_BUTTON_TIME, false);
                break;

                // sleep
                case eMouseArea_Sleep:
                    mainState = eMainSleep;
                    for (int i = 0; i < 4; ++i)
                        game.characters[i].enterSleep();
                break;

                // wake up
                case eMouseArea_WakeUp:
                    wakeUp();
                break;

                // exit pause
                case eMouseArea_ExitPause:
                    mainState = mainStateBeforePause;
                break;

                default:
                break;
            }
        }

        if (mouse.mButtonPressed == true)
        {
            if (clickedArea->type == eMouseArea_Eye)
                isPressingEye = true;
        }
    }

    if (keyboard.keys[CKeyboard::ePause] == true)
    {
        if (mainState != eMainPause)
        {
            mainStateBeforePause = mainState;
            mainState = eMainPause;
        }
        else
        {
            mainState = mainStateBeforePause;
        }
    }

    updateArrows(clickedArea);
    updateMessages();
    updateWeapons();
    updateDiskButtons();
}

//------------------------------------------------------------
void CInterface::wakeUp()
{
    mainState = eMainInventory;
    for (int i = 0; i < 4; ++i)
        game.characters[i].exitSleep();
}

//------------------------------------------------------------
void CInterface::updateWeapons()
{
    // attack highlight
    if (attackHighlightTime.update() == true)
    {
        CObject&    weapon = getWeapon(currentWeapon);
        int         weaponType = weapon.getType();
        CObjects::CObjectInfo   weaponInfo = objects.mObjectInfos[weaponType];
        CAttacks::SCombo&        combo = attacks.getCombo(weaponInfo.combo);
        int         attackType = combo.attacks[currentAttack].type;
        CAttacks::SAttack&       attack = attacks.getAttack(attackType);

        weaponCoolDown[currentWeapon].set(attack.cooldown, true);

        CVec2   frontPos = player.getPosFromLocal(CVec2(0, -1));
        CTile*  frontTile = map.getTile(frontPos);

        if (frontTile != NULL)
        {
            // breakable door
            if (frontTile->getType() == eTileDoor &&
                frontTile->getBoolParam("isBreakable") == true &&
                frontTile->getBoolParam("isBroken") == false &&
                (attackType == ATTACK_CHOP ||
                 attackType == ATTACK_KICK ||
                 attackType == ATTACK_SWING ||
                 attackType == ATTACK_HACK ||
                 attackType == ATTACK_BERZERK ||
                 attackType == ATTACK_BASH)
                )
            {
                frontTile->setBoolParam("isBroken", true);
                weaponCoolDown[currentWeapon].set(60, true);
                sound->play(frontPos, "sound/Wooden_Thud.wav");
            }
            else if (attack.type == "Melee")
            {
                // melee attack: look for a monster's group in front of us
                CVec2   monsterPos = player.pos;

                monsterPos.x += (player.dir == 3) - (player.dir == 1);
                monsterPos.y += (player.dir == 2) - (player.dir == 0);

                int groupIndex = map.findMonsterGroupIndex(monsterPos);

                if (groupIndex != -1)
                {
                    CMonsterGroup2& group = monsters.monsterGroups[groupIndex];

                    // only front players can attack with melee weapon
                    if (player.isChampionInFront(monsterPos, currentWeapon) == true)
                    {
                        // choose a target
                        uint8_t target = chooseTarget(monsterPos);
                        damages = game.characters[currentWeapon].getMeleeDamage(monsterPos, attack);

                        if (damages != 0)
                        {
                            // hit
                            damagesDisplayTime.set(DAMAGES_DISPLAY_TIME, false);
                            weaponsAreaState = eWeaponsAreaDamage;
                            group.hitMonster(monsterPos, target, damages);
                        }
                        else
                        {
                            // miss
                            weaponsAreaState = eWeaponsAreaWeapons;
                        }
                    }
                    else
                    {
                        // can't reach
                        weaponsAreaState = eWeaponsAreaWeapons;
                    }
                }
                else
                {
                    // nothing to hit
                    weaponsAreaState = eWeaponsAreaWeapons;
                }
            }
            else if (attack.type == "Ranged")
            {
                weaponsAreaState = eWeaponsAreaWeapons;
            }
            else if (attack.type == "Spell")
            {
                weaponsAreaState = eWeaponsAreaWeapons;
            }
            else if (attack.type == "Special")
            {
                weaponsAreaState = eWeaponsAreaWeapons;
            }
        }
    }

    // damages timer
    if (damagesDisplayTime.update() == true)
        weaponsAreaState = eWeaponsAreaWeapons;

    // weapons cooldowns
    for (int i = 0; i < 4; ++i)
        weaponCoolDown[i].update();
}


//------------------------------------------------------------
uint8_t CInterface::chooseTarget(CVec2 mapPos)
{
    // list the possible targets
    int groupIndex = map.findMonsterGroupIndex(mapPos);
    CMonsterGroup2& group = monsters.monsterGroups[groupIndex];

    uint8_t possibleTargets[2];
    int nbTargets = 0;

    for (int i = 0; i < 4; ++i)
    {
        CMonster&   monster = group.mMonsters[i];

        if (monster.pos != eMonsterPosNone &&
            group.isMonsterInFront(mapPos, monster.pos) == true)
        {
            possibleTargets[nbTargets++] = i;
        }
    }

    // choose a target
    return possibleTargets[RANDOM(nbTargets)];
}

//------------------------------------------------------------
void CInterface::readTextDB()
{
    QDomDocument doc;

    QFile f("databases/texts.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement dbString = root.firstChild().toElement();

    while(!dbString.isNull())
    {
        if (dbString.tagName() == "text")
        {
            std::string id = dbString.attribute("id").toUtf8().constData();
            std::string text = dbString.text().toUtf8().constData();
            textsDB[id] = text;
        }
        dbString = dbString.nextSibling().toElement();
    }
}

//------------------------------------------------------------
std::string CInterface::getText(std::string id)
{
    return  textsDB[id];
}

//------------------------------------------------------------
void CInterface::enterResurrect(int num)
{
    mainState = eMainInventory;
    isResurrecting = true;
    currentChampion = num;
}

//------------------------------------------------------------
void CInterface::addMessage(std::string text, int champion)
{
    if (mMessages.size() == 4)
        mMessages.pop_back();

    SMessage    newMessage;
    newMessage.text = text;
    newMessage.champion = champion;
    newMessage.time.set(MSG_TIME, false);
    mMessages.push_front(newMessage);
    messagePos = MSG_START_POS;
}

//------------------------------------------------------------
void CInterface::drawMessages(QImage* image)
{
    std::list<SMessage>::iterator it;
    int pos = messagePos;

    for (it = mMessages.begin(); it != mMessages.end(); ++it)
    {
        QColor  color;

        if (it->champion < 4)
        {
            color = QColor(championsColors[it->champion][0],
                           championsColors[it->champion][1],
                           championsColors[it->champion][2]);
        }
        else
        {
            color = MAIN_INTERFACE_COLOR;
        }

        drawText(image, CVec2(0, pos / MSG_POS_FACTOR), eFontStandard, (it->text).c_str(), color);
        pos -= 7 * MSG_POS_FACTOR;
    }
}

//------------------------------------------------------------
void CInterface::updateMessages()
{
    // scrolling
    if (messagePos > MSG_DEST_POS)
        messagePos -= MSG_SCROLL_SPEED;

    if (messagePos < MSG_DEST_POS)
        messagePos = MSG_DEST_POS;

    // decrement times
    std::list<SMessage>::iterator it;

    for (it = mMessages.begin(); it != mMessages.end();)
    {
        if (it->time.update() == true)
            it = mMessages.erase(it);
        else
            ++it;
    }
}

//------------------------------------------------------------
void CInterface::updateDiskButtons()
{
    if (diskButtonTime.update() == true)
    {
        switch (clickedDiskButton)
        {
            case eMouseArea_Save:
            {
                QString fileName = QFileDialog::getSaveFileName(NULL, getText("FILE_DIALOG00").c_str(), "saved_games/", getText("FILE_DIALOG02").c_str());

                if (fileName.isEmpty() == true)
                {
                    mainState = eMainDiskDialog;
                }
                else
                {
                    saveName = fileName.toLocal8Bit().constData();
                    mainState = eMainSavingDialog;
                    dummyDiskTimer.set(2, false);
                }
            }
            break;

            case eMouseArea_Load:
            {
                QString fileName = QFileDialog::getOpenFileName(NULL, getText("FILE_DIALOG01").c_str(), "saved_games/", getText("FILE_DIALOG02").c_str());

                if (fileName.isEmpty() == true)
                {
                    mainState = eMainDiskDialog;
                }
                else
                {
                    saveName = fileName.toLocal8Bit().constData();
                    mainState = eMainLoadingDialog;
                    dummyDiskTimer.set(2, false);
                }
            }
            break;

            case eMouseArea_Quit:
                mainState = eMainQuitConfirm;
            break;

            case eMouseArea_QuitDiskMenu:
                mainState = eMainInventory;
            break;

            case eMouseArea_GameLoaded:
                mainState = eMainGame;
            break;

            case eMouseArea_GameSaved:
            case eMouseArea_CancelQuitConfirm:
                mainState = eMainDiskDialog;
            break;

            case eMouseArea_AcceptQuitConfirm:
                exit(0);
            break;

            default:
            break;
        }
    }
}

//------------------------------------------------------------
void CInterface::setMainState(EMainState state)
{
    mainState = state;
}

//------------------------------------------------------------
bool CInterface::isGamePaused()
{
    return (isInDiskDialog() == true || mainState == eMainPause);
}

//------------------------------------------------------------
bool CInterface::isSleeping()
{
    return (mainState == eMainSleep);
}

//------------------------------------------------------------
bool CInterface::isFalling()
{
    return (mainState == eMainFall);
}

//------------------------------------------------------------
bool CInterface::isResurrectingCharacter(int num)
{
    if (isResurrecting == true && findLastChampion() == num)
        return true;

    return false;
}
